r"""Mean Deviation Similarity Index (MDSI)

This module implements the MDSI in PyTorch.

Original:
    https://www.mathworks.com/matlabcentral/fileexchange/59809-mdsi-ref-dist-combmethod

References:
    .. [Nafchi2016] Mean Deviation Similarity Index: Efficient and Reliable Full-Reference Image Quality Evaluator (Nafchi et al., 2016)
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from .utils import _jit, assert_type, reduce_tensor
from .utils import complex as cx
from .utils.color import ColorConv
from .utils.functional import (
    prewitt_kernel,
    gradient_kernel,
    channel_conv,
    l2_norm,
)


@_jit
def mdsi(
    x: Tensor,
    y: Tensor,
    kernel: Tensor,
    combination: str = 'sum',
    value_range: float = 1.,
    c1: float = 140. / (255. ** 2),
    c2: float = 55. / (255. ** 2),
    c3: float = 550. / (255. ** 2),
    alpha: float = 0.6,  # 'sum'
    beta: float = 0.1,  # 'prod'
    gamma: float = 0.2,  # 'prod'
    rho: float = 1.,
    q: float = 0.25,
    o: float = 0.25,
) -> Tensor:
    r"""Returns the MDSI between :math:`x` and :math:`y`,
    without downsampling and color space conversion.

    Args:
        x: An input tensor, :math:`(N, 3, H, W)`.
        y: A target tensor, :math:`(N, 3, H, W)`.
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
        combination: Specifies the scheme to combine the gradient
            and chromaticity similarities (GS, CS): `'sum'` | `'prod'`.
        value_range: The value range :math:`L` of the inputs (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Nafchi2016]_.

    Returns:
        The MDSI vector, :math:`(N,)`.

    Example:
        >>> x = torch.rand(5, 1, 256, 256)
        >>> y = torch.rand(5, 1, 256, 256)
        >>> kernel = gradient_kernel(prewitt_kernel())
        >>> l = mdsi(x, y, kernel)
        >>> l.size()
        torch.Size([5])
    """

    c1 *= value_range ** 2
    c2 *= value_range ** 2
    c3 *= value_range ** 2

    l_x, hm_x = x[:, :1], x[:, 1:]
    l_y, hm_y = y[:, :1], y[:, 1:]

    # Gradient magnitude
    pad = kernel.size(-1) // 2

    gm_x = l2_norm(channel_conv(l_x, kernel, padding=pad), dims=[1])
    gm_y = l2_norm(channel_conv(l_y, kernel, padding=pad), dims=[1])
    gm_avg = l2_norm(channel_conv(l_x + l_y, kernel, padding=pad), dims=[1]) / 2

    gm_x_sq, gm_y_sq, gm_avg_sq = gm_x ** 2, gm_y ** 2, gm_avg ** 2

    # Gradient similarity
    gs_x_y = (2. * gm_x * gm_y + c1) / (gm_x_sq + gm_y_sq + c1)
    gs_x_avg = (2. * gm_x * gm_avg + c2) / (gm_x_sq + gm_avg_sq + c2)
    gs_y_avg = (2. * gm_y * gm_avg + c2) / (gm_y_sq + gm_avg_sq + c2)

    gs = gs_x_y + gs_x_avg - gs_y_avg

    # Chromaticity similarity
    cs_num = 2. * (hm_x * hm_y).sum(dim=1) + c3
    cs_den = (hm_x ** 2 + hm_y ** 2).sum(dim=1) + c3
    cs = cs_num / cs_den

    # Gradient-chromaticity similarity
    gs = cx.complx(gs, torch.zeros_like(gs))
    cs = cx.complx(cs, torch.zeros_like(cs))

    if combination == 'prod':
        gcs = cx.prod(cx.pow(gs, gamma), cx.pow(cs, beta))
    else:  # combination == 'sum'
        gcs = alpha * gs + (1. - alpha) * cs

    # Mean deviation similarity
    gcs_q = cx.pow(gcs, q)
    gcs_q_avg = gcs_q.mean(dim=(-2, -3), keepdim=True)
    score = cx.mod(gcs_q - gcs_q_avg, squared=True) ** (rho / 2)
    mds = score.mean(dim=(-1, -2)) ** (o / rho)

    return mds


class MDSI(nn.Module):
    r"""Creates a criterion that measures the MDSI
    between an input and a target.

    Before applying :func:`mdsi`, the input and target are converted from
    RBG to LHM and downsampled by a factor :math:`\frac{\min(H, W)}{256}`.

    Args:
        downsample: Whether downsampling is enabled or not.
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
            If `None`, use the Prewitt kernel instead.
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Note:
        `**kwargs` are passed to :func:`mdsi`.

    Shapes:
        input: :math:`(N, 3, H, W)`
        target: :math:`(N, 3, H, W)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = MDSI().cuda()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    def __init__(
        self,
        downsample: bool = True,
        kernel: Tensor = None,
        reduction: str = 'mean',
        **kwargs,
    ):
        super().__init__()

        if kernel is None:
            kernel = gradient_kernel(prewitt_kernel())

        self.register_buffer('kernel', kernel)

        self.convert = ColorConv('RGB', 'LHM')
        self.downsample = downsample
        self.reduction = reduction
        self.value_range = kwargs.get('value_range', 1.)
        self.kwargs = kwargs

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            device=self.kernel.device,
            dim_range=(4, 4),
            n_channels=3,
            value_range=(0., self.value_range),
        )

        # Downsample
        if self.downsample:
            _, _, h, w = input.size()
            M = round(min(h, w) / 256)

            if M > 1:
                input = F.avg_pool2d(input, kernel_size=M, ceil_mode=True)
                target = F.avg_pool2d(target, kernel_size=M, ceil_mode=True)

        # RGB to LHM
        input = self.convert(input)
        target = self.convert(target)

        # MDSI
        l = mdsi(input, target, kernel=self.kernel, **self.kwargs)

        return reduce_tensor(l, self.reduction)



r"""Differentiable and JITable complex tensor API"""

import torch

from torch import Tensor


def complx(real: Tensor, imag: Tensor) -> Tensor:
    r"""Returns a complex tensor with its real part equal to :math:`\Re` and
    its imaginary part equal to :math:`\Im`.

    .. math:: c = \Re + i \Im

    Args:
        real: A tensor :math:`\Re`, :math:`(*,)`.
        imag: A tensor :math:`\Im`, :math:`(*,)`.

    Returns:
        The complex tensor, :math:`(*, 2)`.

    Example:
        >>> x = torch.tensor([2., 0.7071])
        >>> y = torch.tensor([0., 0.7071])
        >>> complx(x, y)
        tensor([[2.0000, 0.0000],
                [0.7071, 0.7071]])
    """

    return torch.stack([real, imag], dim=-1)


def real(x: Tensor) -> Tensor:
    r"""Returns the real part of :math:`x`.

    Args:
        x: A complex tensor, :math:`(*, 2)`.

    Returns:
        The real tensor, :math:`(*,)`.

    Example:
        >>> x = torch.tensor([[2., 0.], [0.7071, 0.7071]])
        >>> real(x)
        tensor([2.0000, 0.7071])
    """

    return x[..., 0]


def imag(x: Tensor) -> Tensor:
    r"""Returns the imaginary part of :math:`x`.

    Args:
        x: A complex tensor, :math:`(*, 2)`.

    Returns:
        The imaginary tensor, :math:`(*,)`.

    Example:
        >>> x = torch.tensor([[2., 0.], [0.7071, 0.7071]])
        >>> imag(x)
        tensor([0.0000, 0.7071])
    """

    return x[..., 1]


def conj(x: Tensor) -> Tensor:
    r"""Returns the element-wise conjugate of :math:`x`.

    .. math:: \bar{x} = \Re(x) - i \Im(x)

    Args:
        x: A complex tensor, :math:`(*, 2)`.

    Returns:
        The conjugate tensor, :math:`(*, 2)`.

    Example:
        >>> x = torch.tensor([[2., 0.], [0.7071, 0.7071]])
        >>> conj(x)
        tensor([[ 2.0000, -0.0000],
                [ 0.7071, -0.7071]])
    """

    return x * torch.tensor([1., -1.]).to(x)


def turn(x: Tensor) -> Tensor:
    r"""Returns the element-wise product of :math:`x` with :math:`i`.

    .. math:: i x = -\Im(x) + i \Re(x)

    Args:
        x: A complex tensor, :math:`(*, 2)`.

    Returns:
        The turned tensor, :math:`(*, 2)`.

    Example:
        >>> x = torch.tensor([[2., 0.], [0.7071, 0.7071]])
        >>> turn(x)
        tensor([[-0.0000,  2.0000],
                [-0.7071,  0.7071]])
    """

    return complx(-imag(x), real(x))


def polar(r: Tensor, phi: Tensor) -> Tensor:
    r"""Returns a complex tensor with its modulus equal to :math:`r`
    and its phase equal to :math:`\phi`.

    .. math:: c = r \exp(i \phi)

    Args:
        r: A tensor :math:`r`, :math:`(*,)`.
        phi: A tensor :math:`\phi`, :math:`(*,)`.

    Returns:
        The complex tensor, :math:`(*, 2)`.

    Example:
        >>> x = torch.tensor([2., 1.])
        >>> y = torch.tensor([0., 0.7854])
        >>> polar(x, y)
        tensor([[2.0000, 0.0000],
                [0.7071, 0.7071]])
    """

    return complx(r * torch.cos(phi), r * torch.sin(phi))


def mod(x: Tensor, squared: bool = False) -> Tensor:
    r"""Returns the modulus (absolute value) of :math:`x`.

    .. math:: \left| x \right| = \sqrt{ \Re(x)^2 + \Im(x)^2 }

    Args:
        x: A complex tensor, :math:`(*, 2)`.
        squared: Whether the output is squared or not.

    Returns:
        The modulus tensor, :math:`(*,)`.

    Example:
        >>> x = torch.tensor([[2., 0.], [0.7071, 0.7071]])
        >>> mod(x)
        tensor([2.0000, 1.0000])
    """

    x = x.square().sum(dim=-1)

    if not squared:
        x = torch.sqrt(x)

    return x


def angle(x: Tensor) -> Tensor:
    r"""Returns the angle (phase) of :math:`x`.

    .. math:: \phi(x) = \operatorname{atan2}(\Im(x), \Re(x))

    Args:
        x: A complex tensor, :math:`(*, 2)`.

    Returns:
        The angle tensor, :math:`(*,)`.

    Example:
        >>> x = torch.tensor([[2., 0.], [0.7071, 0.7071]])
        >>> angle(x)
        tensor([0.0000, 0.7854])
    """

    return torch.atan2(imag(x), real(x))


def prod(x: Tensor, y: Tensor) -> Tensor:
    r"""Returns the element-wise product of :math:`x` and :math:`y`.

    .. math::
        x y = \Re(x) \Re(y) - \Im(x) \Im(y)
            + i \left( \Re(x) \Im(y) - \Im(x) \Re(y) \right)

    Args:
        x: A complex tensor, :math:`(*, 2)`.
        y: A complex tensor, :math:`(*, 2)`.

    Returns:
        The product tensor, :math:`(*, 2)`.

    Example:
        >>> x = torch.tensor([[2.,  0.], [0.7071,  0.7071]])
        >>> y = torch.tensor([[2., -0.], [0.7071, -0.7071]])
        >>> prod(x, y)
        tensor([[4.0000, 0.0000],
                [1.0000, 0.0000]])
    """

    x_r, x_i = real(x), imag(x)
    y_r, y_i = real(y), imag(y)

    return complx(x_r * y_r - x_i * y_i, x_i * y_r + x_r * y_i)


def dot(x: Tensor, y: Tensor) -> Tensor:
    r"""Returns the element-wise dot-product of :math:`x` and :math:`y`.

    .. math:: x \odot y = \Re(x) \Re(y) + \Im(x) \Im(y)

    Args:
        x: A complex tensor, :math:`(*, 2)`.
        y: A complex tensor, :math:`(*, 2)`.

    Returns:
        The dot-product tensor, :math:`(*,)`.

    Example:
        >>> x = torch.tensor([[2.,  0.], [0.7071,  0.7071]])
        >>> y = torch.tensor([[2., -0.], [0.7071, -0.7071]])
        >>> dot(x, y)
        tensor([4., 0.])
    """

    return (x * y).sum(dim=-1)


def pow(x: Tensor, exponent: float) -> Tensor:
    r"""Returns the power of :math:`x` with `exponent`.

    .. math:: x^p = \left| x \right|^p \exp(i \phi(x))^p

    Args:
        x: A complex tensor, :math:`(*, 2)`.
        exponent: The exponent :math:`p`.

    Returns:
        The power tensor, :math:`(*, 2)`.

    Example:
        >>> x = torch.tensor([[2., 0.], [0.7071, 0.7071]])
        >>> pow(x, 2.)
        tensor([[ 4.0000e+00,  0.0000e+00],
                [-4.3711e-08,  9.9998e-01]])
    """

    r = mod(x, squared=True) ** (exponent / 2)
    phi = angle(x) * exponent

    return polar(r, phi)



r"""Peak Signal-to-Noise Ratio (PSNR)

This module implements the PSNR in PyTorch.

Wikipedia:
    https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio
"""

import torch
import torch.nn as nn

from torch import Tensor

from .utils import _jit, assert_type, reduce_tensor


@_jit
def mse(x: Tensor, y: Tensor) -> Tensor:
    r"""Returns the Mean Squared Error (MSE) between :math:`x` and :math:`y`.

    .. math::
        \text{MSE}(x, y) = \frac{1}{\text{size}(x)} \sum_i (x_i - y_i)^2

    Args:
        x: An input tensor, :math:`(N, *)`.
        y: A target tensor, :math:`(N, *)`.

    Returns:
        The MSE vector, :math:`(N,)`.

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> y = torch.rand(5, 3, 256, 256)
        >>> l = mse(x, y)
        >>> l.size()
        torch.Size([5])
    """

    return ((x - y) ** 2).view(x.size(0), -1).mean(dim=-1)


@_jit
def psnr(
    x: Tensor,
    y: Tensor,
    epsilon: float = 1e-8,
    value_range: float = 1.,
) -> Tensor:
    r"""Returns the PSNR between :math:`x` and :math:`y`.

    .. math::
        \text{PSNR}(x, y) = 10 \log_{10} \left( \frac{L^2}{\text{MSE}(x, y)} \right)

    Args:
        x: An input tensor, :math:`(N, *)`.
        y: A target tensor, :math:`(N, *)`.
        epsilon: A numerical stability term.
        value_range: The value range :math:`L` of the inputs (usually `1.` or `255`).

    Returns:
        The PSNR vector, :math:`(N,)`.

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> y = torch.rand(5, 3, 256, 256)
        >>> l = psnr(x, y)
        >>> l.size()
        torch.Size([5])
    """

    return 10 * torch.log10(value_range ** 2 / (mse(x, y) + epsilon))


class PSNR(nn.Module):
    r"""Creates a criterion that measures the PSNR
    between an input and a target.

    Args:
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Note:
        `**kwargs` are passed to :func:`psnr`.

    Shapes:
        input: :math:`(N, *)`
        target: :math:`(N, *)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = PSNR()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = -criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    def __init__(self, reduction: str = 'mean', **kwargs):
        super().__init__()

        self.reduction = reduction
        self.value_range = kwargs.get('value_range', 1.)
        self.kwargs = kwargs

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            dim_range=(1, -1),
            value_range=(0., self.value_range),
        )

        l = psnr(input, target, **self.kwargs)

        return reduce_tensor(l, self.reduction)



r"""Visual Saliency-based Index (VSI)

This module implements the VSI in PyTorch.

Original:
    https://sse.tongji.edu.cn/linzhang/IQA/VSI/VSI.htm

Wikipedia:
    https://en.wikipedia.org/wiki/Salience_(neuroscience)#Visual_saliency_modeling

References:
    .. [Zhang2014] VSI: A Visual Saliency-Induced Index for Perceptual Image Quality Assessment (Zhang et al., 2014)

    .. [Zhang2013] SDSP: A novel saliency detection method by combining simple priors (Zhang et al., 2013)
"""

import torch
import torch.fft as fft
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from .utils import _jit, assert_type, reduce_tensor
from .utils import complex as cx
from .utils.color import ColorConv, rgb_to_xyz, xyz_to_lab
from .utils.functional import (
    scharr_kernel,
    gradient_kernel,
    filter_grid,
    log_gabor,
    channel_conv,
    l2_norm,
)


@_jit
def vsi(
    x: Tensor,
    y: Tensor,
    vs_x: Tensor,
    vs_y: Tensor,
    kernel: Tensor,
    value_range: float = 1.,
    c1: float = 1.27,
    c2: float = 386. / (255. ** 2),
    c3: float = 130. / (255. ** 2),
    alpha: float = 0.4,
    beta: float = 0.02,
) -> Tensor:
    r"""Returns the VSI between :math:`x` and :math:`y`,
    without downsampling and color space conversion.

    Args:
        x: An input tensor, :math:`(N, 3 \text{ or } 1, H, W)`.
        y: A target tensor, :math:`(N, 3 \text{ or } 1, H, W)`.
        vs_x: The input visual saliency, :math:`(N, H, W)`.
        vs_y: The target visual saliency, :math:`(N, H, W)`.
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
        value_range: The value range :math:`L` of the inputs (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Zhang2014]_.

    Returns:
        The VSI vector, :math:`(N,)`.

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> y = torch.rand(5, 3, 256, 256)
        >>> filtr = sdsp_filter(x)
        >>> vs_x, vs_y = sdsp(x, filtr), sdsp(y, filtr)
        >>> kernel = gradient_kernel(scharr_kernel())
        >>> l = vsi(x, y, vs_x, vs_y, kernel)
        >>> l.size()
        torch.Size([5])
    """

    c2 *= value_range ** 2
    c3 *= value_range ** 2

    l_x, l_y = x[:, :1], y[:, :1]

    # Visual saliency similarity
    vs_m = torch.max(vs_x, vs_y)
    s_vs = (2 * vs_x * vs_y + c1) / (vs_x ** 2 + vs_y ** 2 + c1)

    # Gradient magnitude similarity
    pad = kernel.size(-1) // 2

    g_x = l2_norm(channel_conv(l_x, kernel, padding=pad), dims=[1])
    g_y = l2_norm(channel_conv(l_y, kernel, padding=pad), dims=[1])

    s_g = (2 * g_x * g_y + c2) / (g_x ** 2 + g_y ** 2 + c2)

    # Chorminance similarity
    if x.size(1) == 3:
        mn_x, mn_y = x[:, 1:], y[:, 1:]

        s_c = (2 * mn_x * mn_y + c3) / (mn_x ** 2 + mn_y ** 2 + c3)
        s_c = s_c.prod(dim=1)

        s_c = cx.complx(s_c, torch.zeros_like(s_c))
        s_c_beta = cx.real(cx.pow(s_c, beta))

        s_vs = s_vs * s_c_beta

    # Visual Saliency-based Index
    s = s_vs * s_g ** alpha
    vsi = (s * vs_m).sum(dim=(-1, -2)) / vs_m.sum(dim=(-1, -2))

    return vsi


@_jit
def sdsp_filter(
    x: Tensor,
    omega_0: float = 0.021,
    sigma_f: float = 1.34,
) -> Tensor:
    r"""Returns the log-Gabor filter for :func:`sdsp`.

    Args:
        x: An input tensor, :math:`(*, H, W)`.

    Note:
        For the remaining arguments, refer to [Zhang2013]_.

    Returns:
        The filter tensor, :math:`(H, W)`.
    """

    r, _ = filter_grid(x)
    filtr = log_gabor(r, omega_0, sigma_f)
    filtr = filtr * (r <= 0.5)  # low-pass filter

    return filtr


@_jit
def sdsp(
    x: Tensor,
    filtr: Tensor,
    value_range: float = 1.,
    sigma_c: float = 0.001,
    sigma_d: float = 145.,
) -> Tensor:
    r"""Detects salient regions from :math:`x`.

    Args:
        x: An input tensor, :math:`(N, 3, H, W)`.
        filtr: The frequency domain filter, :math:`(H, W)`.
        value_range: The value range :math:`L` of the input (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Zhang2013]_.

    Returns:
        The visual saliency tensor, :math:`(N, H, W)`.

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> filtr = sdsp_filter(x)
        >>> vs = sdsp(x, filtr)
        >>> vs.size()
        torch.Size([5, 256, 256])
    """

    x_lab = xyz_to_lab(rgb_to_xyz(x, value_range))

    # Frequency prior
    x_f = fft.ifft2(fft.fft2(x_lab) * filtr)
    x_f = cx.real(torch.view_as_real(x_f))

    s_f = l2_norm(x_f, dims=[1])

    # Color prior
    x_ab = x_lab[:, 1:]

    lo, _ = x_ab.flatten(-2).min(dim=-1)
    up, _ = x_ab.flatten(-2).max(dim=-1)

    lo = lo.view(lo.shape + (1, 1))
    up = up.view(lo.shape)
    span = torch.where(up > lo, up - lo, torch.tensor(1.).to(lo))

    x_ab = (x_ab - lo) / span

    s_c = 1. - torch.exp(-torch.sum(x_ab ** 2, dim=1) / sigma_c ** 2)

    # Location prior
    a, b = [
        torch.arange(n).to(x) - (n - 1) / 2
        for n in x.shape[-2:]
    ]

    s_d = torch.exp(-(a[None, :] ** 2 + b[:, None] ** 2) / sigma_d ** 2)

    # Visual saliency
    vs = s_f * s_c * s_d

    return vs


class VSI(nn.Module):
    r"""Creates a criterion that measures the VSI
    between an input and a target.

    Before applying :func:`vsi`, the input and target are converted from
    RBG to L(MN) and downsampled by a factor :math:`\frac{\min(H, W)}{256}`.

    The visual saliency maps of the input and target are determined by :func:`sdsp`.

    Args:
        chromatic: Whether to use the chromatic channels (MN) or not.
        downsample: Whether downsampling is enabled or not.
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
            If `None`, use the Scharr kernel instead.
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Note:
        `**kwargs` are passed to :func:`vsi`.

    Shapes:
        input: :math:`(N, 3, H, W)`
        target: :math:`(N, 3, H, W)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = VSI().cuda()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = 1 - criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    def __init__(
        self,
        chromatic: bool = True,
        downsample: bool = True,
        kernel: Tensor = None,
        reduction: str = 'mean',
        **kwargs,
    ):
        super().__init__()

        if kernel is None:
            kernel = gradient_kernel(scharr_kernel())

        self.register_buffer('kernel', kernel)
        self.register_buffer('filter', torch.zeros((0, 0)))

        self.convert = ColorConv('RGB', 'LMN' if chromatic else 'L')
        self.downsample = downsample
        self.reduction = reduction
        self.value_range = kwargs.get('value_range', 1.)
        self.kwargs = kwargs

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            device=self.kernel.device,
            dim_range=(4, 4),
            n_channels=3,
            value_range=(0., self.value_range),
        )

        # Downsample
        if self.downsample:
            _, _, h, w = input.size()
            M = round(min(h, w) / 256)

            if M > 1:
                input = F.avg_pool2d(input, kernel_size=M, ceil_mode=True)
                target = F.avg_pool2d(target, kernel_size=M, ceil_mode=True)

        # Visual saliency
        if self.filter.shape != input.shape[-2:]:
            self.filter = sdsp_filter(input)

        vs_input = sdsp(input, self.filter, self.value_range)
        vs_target = sdsp(target, self.filter, self.value_range)

        # RGB to L(MN)
        input = self.convert(input)
        target = self.convert(target)

        # VSI
        l = vsi(input, target, vs_input, vs_target, kernel=self.kernel, **self.kwargs)

        return reduce_tensor(l, self.reduction)



r"""Gradient Magnitude Similarity Deviation (GMSD)
and Multi-Scale Gradient Magnitude Similarity Deviation (MS-GMSD)

This module implements the GMSD and MS-GMSD in PyTorch.

Original:
    https://www4.comp.polyu.edu.hk/~cslzhang/IQA/GMSD/GMSD.htm

References:
    .. [Xue2013] Gradient Magnitude Similarity Deviation: An Highly Efficient Perceptual Image Quality Index (Xue et al., 2013)

    .. [Zhang2017] Gradient Magnitude Similarity Deviation on multiple scales for color image quality assessment (Zhang et al., 2017)
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from .utils import _jit, assert_type, reduce_tensor
from .utils.color import ColorConv
from .utils.functional import (
    prewitt_kernel,
    gradient_kernel,
    channel_conv,
    l2_norm,
)


@_jit
def gmsd(
    x: Tensor,
    y: Tensor,
    kernel: Tensor,
    value_range: float = 1.,
    c: float = 170. / (255. ** 2),
    alpha: float = 0.,
) -> Tensor:
    r"""Returns the GMSD between :math:`x` and :math:`y`,
    without color space conversion and downsampling.

    :math:`\text{GMSD}(x, y)` is the standard deviation of the Gradient
    Magnitude Similarity (GMS).

    .. math::
        \text{GMS}(x, y) &= \frac{(2 - \alpha) \text{GM}(x) \text{GM}(y) + C}
            {\text{GM}(x)^2 + \text{GM}(y)^2 - \alpha \text{GM}(x) \text{GM}(y) + C} \\
        \text{GM}(z) &= \left\| \nabla z \right\|_2

    where :math:`\nabla z` is the result of a gradient convolution over :math:`z`.

    Args:
        x: An input tensor, :math:`(N, 1, H, W)`.
        y: A target tensor, :math:`(N, 1, H, W)`.
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
        value_range: The value range :math:`L` of the inputs (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Xue2013]_.

    Returns:
        The GMSD vector, :math:`(N,)`.

    Example:
        >>> x = torch.rand(5, 1, 256, 256)
        >>> y = torch.rand(5, 1, 256, 256)
        >>> kernel = gradient_kernel(prewitt_kernel())
        >>> l = gmsd(x, y, kernel)
        >>> l.size()
        torch.Size([5])
    """

    c *= value_range ** 2

    # Gradient magnitude
    pad = kernel.size(-1) // 2

    gm_x = l2_norm(channel_conv(x, kernel, padding=pad), dims=[1])
    gm_y = l2_norm(channel_conv(y, kernel, padding=pad), dims=[1])

    gm_xy = gm_x * gm_y

    # Gradient magnitude similarity
    gms_num = (2. - alpha) * gm_xy + c
    gms_den = gm_x ** 2 + gm_y ** 2 + c

    if alpha > 0.:
        gms_den = gms_den - alpha * gm_xy

    gms = gms_num / gms_den

    # Gradient magnitude similarity deviation
    gmsd = torch.std(gms, dim=(-1, -2))

    return gmsd


@_jit
def ms_gmsd(
    x: Tensor,
    y: Tensor,
    kernel: Tensor,
    weights: Tensor,
    value_range: float = 1.,
    c: float = 170. / (255. ** 2),
    alpha: float = 0.5,
) -> Tensor:
    r"""Returns the MS-GMSD between :math:`x` and :math:`y`,
    without color space conversion.

    .. math::
        \text{MS-GMSD}(x, y) =
            \sqrt{\sum^{M}_{i = 1} w_i \text{GMSD}(x^i, y^i) ** 2}

    where :math:`x^i` and :math:`y^i` are obtained by downsampling
    the initial tensors by a factor :math:`2^{i - 1}`.

    Args:
        x: An input tensor, :math:`(N, 1, H, W)`.
        y: A target tensor, :math:`(N, 1, H, W)`.
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
        weights: The weights :math:`w_i` of the scales, :math:`(M,)`.
        value_range: The value range :math:`L` of the inputs (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Zhang2017]_.

    Returns:
        The MS-GMSD vector, :math:`(N,)`.

    Example:
        >>> x = torch.rand(5, 1, 256, 256)
        >>> y = torch.rand(5, 1, 256, 256)
        >>> kernel = gradient_kernel(prewitt_kernel())
        >>> weights = torch.rand(4)
        >>> l = ms_gmsd(x, y, kernel, weights)
        >>> l.size()
        torch.Size([5])
    """

    gmsds = []

    for i in range(weights.numel()):
        if i > 0:
            x = F.avg_pool2d(x, kernel_size=2, ceil_mode=True)
            y = F.avg_pool2d(y, kernel_size=2, ceil_mode=True)

        gmsds.append(gmsd(
            x, y, kernel,
            value_range=value_range,
            c=c, alpha=alpha,
        ))

    msgmsd = weights * torch.stack(gmsds, dim=-1) ** 2
    msgmsd = msgmsd.sum(dim=-1).sqrt()

    return msgmsd


class GMSD(nn.Module):
    r"""Creates a criterion that measures the GMSD
    between an input and a target.

    Before applying :func:`gmsd`, the input and target are converted from
    RBG to Y, the luminance color space, and downsampled by a factor 2.

    Args:
        downsample: Whether downsampling is enabled or not.
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
            If `None`, use the Prewitt kernel instead.
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.


    Note:
        `**kwargs` are passed to :func:`gmsd`.

    Shapes:
        input: :math:`(N, 3, H, W)`
        target: :math:`(N, 3, H, W)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = GMSD().cuda()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    def __init__(
        self,
        downsample: bool = True,
        kernel: Tensor = None,
        reduction: str = 'mean',
        **kwargs,
    ):
        super().__init__()

        if kernel is None:
            kernel = gradient_kernel(prewitt_kernel())

        self.register_buffer('kernel', kernel)

        self.convert = ColorConv('RGB', 'Y')
        self.downsample = downsample
        self.reduction = reduction
        self.value_range = kwargs.get('value_range', 1.)
        self.kwargs = kwargs

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            device=self.kernel.device,
            dim_range=(4, 4),
            n_channels=3,
            value_range=(0., self.value_range),
        )

        # Downsample
        if self.downsample:
            input = F.avg_pool2d(input, 2, ceil_mode=True)
            target = F.avg_pool2d(target, 2, ceil_mode=True)

        # RGB to Y
        input = self.convert(input)
        target = self.convert(target)

        # GMSD
        l = gmsd(input, target, kernel=self.kernel, **self.kwargs)

        return reduce_tensor(l, self.reduction)


class MS_GMSD(nn.Module):
    r"""Creates a criterion that measures the MS-GMSD
    between an input and a target.

    Before applying :func:`ms_gmsd`, the input and target are converted from
    RBG to Y, the luminance color space.

    Args:
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
            If `None`, use the Prewitt kernel instead.
        weights: The weights of the scales, :math:`(M,)`.
            If `None`, use the :const:`MS_GMSD.WEIGHTS` instead.
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Note:
        `**kwargs` are passed to :func:`ms_gmsd`.

    Shapes:
        input: :math:`(N, 3, H, W)`
        target: :math:`(N, 3, H, W)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = MS_GMSD().cuda()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    WEIGHTS: Tensor = torch.tensor([0.096, 0.596, 0.289, 0.019])
    r"""Scale weights of [Zhang2017]_."""

    def __init__(
        self,
        kernel: Tensor = None,
        weights: Tensor = None,
        reduction: str = 'mean',
        **kwargs,
    ):
        super().__init__()

        if kernel is None:
            kernel = gradient_kernel(prewitt_kernel())

        if weights is None:
            weights = self.WEIGHTS

        self.register_buffer('kernel', kernel)
        self.register_buffer('weights', weights)

        self.convert = ColorConv('RGB', 'Y')
        self.reduction = reduction
        self.value_range = kwargs.get('value_range', 1.)
        self.kwargs = kwargs

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            device=self.kernel.device,
            dim_range=(4, 4),
            n_channels=3,
            value_range=(0., self.value_range),
        )

        # RGB to Y
        input = self.convert(input)
        target = self.convert(target)

        # MS-GMSD
        l = ms_gmsd(
            input,
            target,
            kernel=self.kernel,
            weights=self.weights,
            **self.kwargs,
        )

        return reduce_tensor(l, self.reduction)



r"""Total Variation (TV)

This module implements the TV in PyTorch.

Wikipedia:
    https://en.wikipedia.org/wiki/Total_variation
"""

import torch
import torch.nn as nn

from torch import Tensor

from .utils import _jit, assert_type, reduce_tensor


@_jit
def tv(x: Tensor, norm: str = 'L1') -> Tensor:
    r"""Returns the TV of :math:`x`.

    With `'L1'`,

    .. math::
        \text{TV}(x) = \sum_{i, j}
            \left| x_{i+1, j} - x_{i, j} \right| +
            \left| x_{i, j+1} - x_{i, j} \right|

    Alternatively, with `'L2'`,

    .. math::
        \text{TV}(x) = \left( \sum_{c, i, j}
            (x_{c, i+1, j} - x_{c, i, j})^2 +
            (x_{c, i, j+1} - x_{c, i, j})^2 \right)^{\frac{1}{2}}

    Args:
        x: An input tensor, :math:`(*, C, H, W)`.
        norm: Specifies the norm funcion to apply:
            `'L1'` | `'L2'` | `'L2_squared'`.

    Returns:
        The TV tensor, :math:`(*,)`.

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> l = tv(x)
        >>> l.size()
        torch.Size([5])
    """

    w_var = torch.diff(x, dim=-1)
    h_var = torch.diff(x, dim=-2)

    if norm == 'L1':
        w_var = w_var.abs()
        h_var = h_var.abs()
    else:  # norm in ['L2', 'L2_squared']
        w_var = w_var ** 2
        h_var = h_var ** 2

    var = w_var.sum(dim=(-1, -2, -3)) + h_var.sum(dim=(-1, -2, -3))

    if norm == 'L2':
        var = torch.sqrt(var)

    return var


class TV(nn.Module):
    r"""Creates a criterion that measures the TV of an input.

    Args:
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Note:
        `**kwargs` are passed to :func:`tv`.

    Shapes:
        input: :math:`(*, C, H, W)`
        output: :math:`(*,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = TV()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> l = criterion(x)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    def __init__(self, reduction: str = 'mean', **kwargs):
        super().__init__()

        self.reduction = reduction
        self.kwargs = kwargs

    def forward(self, input: Tensor) -> Tensor:
        assert_type(input, dim_range=(3, -1))

        l = tv(input, **self.kwargs)

        return reduce_tensor(l, self.reduction)



r"""Miscellaneous tools and general purpose components"""

import os
import torch

from torch import Tensor
from typing import List, Tuple


if os.getenv('PIQA_JIT') == '1':
    _jit = torch.jit.script
else:
    _jit = lambda f: f


__piqa_debug__ = __debug__

def set_debug(mode: bool = False) -> bool:
    r"""Sets and returns whether debugging is enabled or not.
    If `__debug__` is `False`, this function has not effect.

    Example:
        >>> set_debug(False)
        False
    """

    global __piqa_debug__

    __piqa_debug__ = __debug__ and mode

    return __piqa_debug__


def broadcastable(*shapes) -> bool:
    r"""Returns whether `shapes` are broadcastable.

    Example:
        >>> x = torch.rand(3, 2, 1)
        >>> y = torch.rand(1, 2, 3)
        >>> z = torch.rand(2, 2, 2)
        >>> broadcastable(x.shape, y.shape)
        True
        >>> broadcastable(y.shape, z.shape)
        False
    """

    try:
        torch.broadcast_shapes(*shapes)
    except RuntimeError as e:
        return False
    else:
        return True


def assert_type(
    *tensors,
    device: torch.device = None,
    dim_range: Tuple[int, int] = None,
    n_channels: int = None,
    value_range: Tuple[float, float] = None,
) -> None:
    r"""Asserts that the types, devices, shapes and values of `tensors` are
    valid with respect to some requirements.

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> y = torch.rand(5, 3, 256, 256)
        >>> assert_type(x, y, dim_range=(4, 4), n_channels=3)
    """

    if not __piqa_debug__:
        return

    if device is None:
        device = tensors[0].device

    shapes = [tuple(t.shape) for t in tensors]
    assert broadcastable(*shapes), f"Expected all tensors to have broadcastable shapes, but got {shapes}."

    for t in tensors:
        assert t.device == device, f"Expected all tensors to be on the same device, but got {str(t.device)} and {str(device)}."

        if dim_range is None:
            pass
        elif dim_range[0] == dim_range[1]:
            assert t.dim() == dim_range[0], f"Expected number of dimensions to be ' {dim_range[0]}, but got {t.dim()}."
        elif dim_range[0] < dim_range[1]:
            assert dim_range[0] <= t.dim(), f"Expected number of dimensions to be greater or equal to {dim_range[0]}, but got {t.dim()}."
            assert t.dim() <= dim_range[1], f"Expected number of dimensions to be lower or equal to {dim_range[1]}, but got {t.dim()}."
        else:
            assert dim_range[0] <= t.dim(), f"Expected number of dimensions to be greater or equal to {dim_range[0]}, but got {t.dim()}."

        if n_channels is not None:
            assert t.size(1) == n_channels, f"Expected number of channels to be {n_channels}, but got {t.size(1)}."

        if value_range is not None:
            assert value_range[0] <= t.min(), f"Expected all values to be greater or equal to {value_range[0]}, but got {t.min().item()}."
            assert t.max() <= value_range[1], f"Expected all values to be lower or equal to {value_range[1]}, but got {t.max().item()}."


@_jit
def reduce_tensor(x: Tensor, reduction: str = 'mean') -> Tensor:
    r"""Returns the reduction of :math:`x`.

    Args:
        x: A tensor, :math:`(*,)`.
        reduction: Specifies the reduction type:
            `'none'` | `'mean'` | `'sum'`.

    Example:
        >>> x = torch.arange(5)
        >>> reduce_tensor(x, reduction='sum')
        tensor(10)
    """

    if reduction == 'mean':
        return x.mean()
    elif reduction == 'sum':
        return x.sum()

    return x



r"""Structural Similarity (SSIM) and Multi-Scale Structural Similarity (MS-SSIM)

This module implements the SSIM and MS-SSIM in PyTorch.

Original:
    https://ece.uwaterloo.ca/~z70wang/research/ssim/

Wikipedia:
    https://en.wikipedia.org/wiki/Structural_similarity

References:
    .. [Wang2004a] Image quality assessment: From error visibility to structural similarity (Wang et al., 2004)

    .. [Wang2004b] Multiscale structural similarity for image quality assessment (Wang et al., 2004)
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Tuple

from .utils import _jit, assert_type, reduce_tensor
from .utils.functional import (
    gaussian_kernel,
    kernel_views,
    channel_convs,
)


@_jit
def ssim(
    x: Tensor,
    y: Tensor,
    kernel: Tensor,
    channel_avg: bool = True,
    padding: bool = False,
    value_range: float = 1.,
    k1: float = 0.01,
    k2: float = 0.03,
) -> Tuple[Tensor, Tensor]:
    r"""Returns the SSIM and Contrast Sensitivity (CS) between
    :math:`x` and :math:`y`.

    .. math::
        \text{SSIM}(x, y) &=
            \frac{2 \mu_x \mu_y + C_1}{\mu^2_x + \mu^2_y + C_1} \text{CS}(x, y) \\
        \text{CS}(x, y) &=
            \frac{2 \sigma_{xy} + C_2}{\sigma^2_x + \sigma^2_y + C_2}

    where :math:`\mu_x`, :math:`\mu_y`, :math:`\sigma^2_x`, :math:`\sigma^2_y` and
    :math:`\sigma_{xy}` are the results of a smoothing convolution over
    :math:`x`, :math:`y`, :math:`(x - \mu_x)^2`, :math:`(y - \mu_y)^2` and
    :math:`(x - \mu_x)(y - \mu_y)`, respectively.

    In practice, SSIM and CS are averaged over the spatial dimensions.
    If `channel_avg` is `True`, they are also averaged over the channels.

    Note:
        The number of spatial dimensions is not limited to 2. You can also apply
        :func:`ssim` (and :class:`SSIM`) on images with 3, or even more, dimensions.

    Args:
        x: An input tensor, :math:`(N, C, H, *)`.
        y: A target tensor, :math:`(N, C, H, *)`.
        kernel: A smoothing kernel, :math:`(C, 1, K)`.
        channel_avg: Whether to average over the channels or not.
        padding: Whether to pad with :math:`\frac{K}{2}` zeros the spatial
            dimensions or not.
        value_range: The value range :math:`L` of the inputs (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Wang2004a]_.

    Returns:
        The SSIM and CS tensors, both :math:`(N, C)` or :math:`(N,)`
        depending on `channel_avg`.

    Example:
        >>> x = torch.rand(5, 3, 64, 64, 64)
        >>> y = torch.rand(5, 3, 64, 64, 64)
        >>> kernel = gaussian_kernel(7).repeat(3, 1, 1)
        >>> ss, cs = ssim(x, y, kernel)
        >>> ss.size(), cs.size()
        (torch.Size([5]), torch.Size([5]))
    """

    c1 = (k1 * value_range) ** 2
    c2 = (k2 * value_range) ** 2

    window = kernel_views(kernel, x.dim() - 2)

    if padding:
        pad = kernel.size(-1) // 2
    else:
        pad = 0

    # Mean (mu)
    mu_x = channel_convs(x, window, pad)
    mu_y = channel_convs(y, window, pad)

    mu_xx = mu_x ** 2
    mu_yy = mu_y ** 2
    mu_xy = mu_x * mu_y

    # Variance (sigma)
    sigma_xx = channel_convs(x ** 2, window, pad) - mu_xx
    sigma_yy = channel_convs(y ** 2, window, pad) - mu_yy
    sigma_xy = channel_convs(x * y, window, pad) - mu_xy

    # Contrast sensitivity (CS)
    cs = (2. * sigma_xy + c2) / (sigma_xx + sigma_yy + c2)

    # Structural similarity (SSIM)
    ss = (2. * mu_xy + c1) / (mu_xx + mu_yy + c1) * cs

    # Average
    if channel_avg:
        ss, cs = ss.flatten(1), cs.flatten(1)
    else:
        ss, cs = ss.flatten(2), cs.flatten(2)

    return ss.mean(dim=-1), cs.mean(dim=-1)


@_jit
def ms_ssim(
    x: Tensor,
    y: Tensor,
    kernel: Tensor,
    weights: Tensor,
    padding: bool = False,
    value_range: float = 1.,
    k1: float = 0.01,
    k2: float = 0.03,
) -> Tensor:
    r"""Returns the MS-SSIM between :math:`x` and :math:`y`.

    .. math::
        \text{MS-SSIM}(x, y) = \text{SSIM}(x^M, y^M)^{\gamma_M}
            \prod^{M - 1}_{i = 1} \text{CS}(x^i, y^i)^{\gamma_i}

    where :math:`x^i` and :math:`y^i` are obtained by downsampling
    the initial tensors by a factor :math:`2^{i - 1}`.

    Args:
        x: An input tensor, :math:`(N, C, H, W)`.
        y: A target tensor, :math:`(N, C, H, W)`.
        kernel: A smoothing kernel, :math:`(C, 1, K)`.
        weights: The weights :math:`\gamma_i` of the scales, :math:`(M,)`.
        padding: Whether to pad with :math:`\frac{K}{2}` zeros the spatial
            dimensions or not.
        value_range: The value range :math:`L` of the inputs (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Wang2004b]_.

    Returns:
        The MS-SSIM vector, :math:`(N,)`.

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> y = torch.rand(5, 3, 256, 256)
        >>> kernel = gaussian_kernel(7).repeat(3, 1, 1)
        >>> weights = torch.rand(5)
        >>> l = ms_ssim(x, y, kernel, weights)
        >>> l.size()
        torch.Size([5])
    """

    css = []

    m = weights.numel()
    for i in range(m):
        if i > 0:
            x = F.avg_pool2d(x, kernel_size=2, ceil_mode=True)
            y = F.avg_pool2d(y, kernel_size=2, ceil_mode=True)

        ss, cs = ssim(
            x, y, kernel,
            channel_avg=False,
            padding=padding,
            value_range=value_range,
            k1=k1, k2=k2,
        )

        css.append(torch.relu(cs) if i + 1 < m else torch.relu(ss))

    msss = torch.stack(css, dim=-1) ** weights
    msss = msss.prod(dim=-1).mean(dim=-1)

    return msss


class SSIM(nn.Module):
    r"""Creates a criterion that measures the SSIM
    between an input and a target.

    Args:
        window_size: The size of the window.
        sigma: The standard deviation of the window.
        n_channels: The number of channels :math:`C`.
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Note:
        `**kwargs` are passed to :func:`ssim`.

    Shapes:
        input: :math:`(N, C, H, *)`
        target: :math:`(N, C, H, *)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = SSIM().cuda()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = 1 - criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    def __init__(
        self,
        window_size: int = 11,
        sigma: float = 1.5,
        n_channels: int = 3,
        reduction: str = 'mean',
        **kwargs,
    ):
        super().__init__()

        kernel = gaussian_kernel(window_size, sigma)

        self.register_buffer('kernel', kernel.repeat(n_channels, 1, 1))

        self.reduction = reduction
        self.value_range = kwargs.get('value_range', 1.)
        self.kwargs = kwargs

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            device=self.kernel.device,
            dim_range=(3, 5),
            n_channels=self.kernel.size(0),
            value_range=(0., self.value_range),
        )

        l = ssim(input, target, kernel=self.kernel, **self.kwargs)[0]

        return reduce_tensor(l, self.reduction)


class MS_SSIM(nn.Module):
    r"""Creates a criterion that measures the MS-SSIM
    between an input and a target.

    Args:
        window_size: The size of the window.
        sigma: The standard deviation of the window.
        n_channels: The number of channels :math:`C`.
        weights: The weights of the scales, :math:`(M,)`.
            If `None`, use :const:`MS_SSIM.WEIGHTS` instead.
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Note:
        `**kwargs` are passed to :func:`ms_ssim`.

    Shapes:
        input: :math:`(N, C, H, W)`
        target: :math:`(N, C, H, W)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = MS_SSIM().cuda()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = 1 - criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    WEIGHTS: Tensor = torch.tensor([0.0448, 0.2856, 0.3001, 0.2363, 0.1333])
    r"""Scale weights of [Wang2004b]_."""

    def __init__(
        self,
        window_size: int = 11,
        sigma: float = 1.5,
        n_channels: int = 3,
        weights: Tensor = None,
        reduction: str = 'mean',
        **kwargs,
    ):
        super().__init__()

        kernel = gaussian_kernel(window_size, sigma)

        self.register_buffer('kernel', kernel.repeat(n_channels, 1, 1))

        if weights is None:
            weights = self.WEIGHTS

        self.register_buffer('weights', weights)

        self.reduction = reduction
        self.value_range = kwargs.get('value_range', 1.)
        self.kwargs = kwargs

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            device=self.kernel.device,
            dim_range=(4, 4),
            n_channels=self.kernel.size(0),
            value_range=(0., self.value_range),
        )

        l = ms_ssim(
            input, target,
            kernel=self.kernel,
            weights=self.weights,
            **self.kwargs,
        )

        return reduce_tensor(l, self.reduction)



r"""Color space conversion tools"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Tuple


def spatial(x: Tensor) -> int:
    r"""Returns the number of spatial dimensions of :math:`x`.

    Args:
        x: A tensor, :math:`(N, C, *)`.
    """

    return len(x.shape) - 2


def color_conv(
    x: Tensor,
    weight: Tensor,
) -> Tensor:
    r"""Returns the color convolution of :math:`x` with the kernel `weight`.

    Args:
        x: A tensor, :math:`(N, C, *)`.
        weight: A weight kernel, :math:`(C', C)`.
    """

    return F.linear(x.transpose(1, -1), weight).transpose(1, -1)


RGB_TO_YIQ = torch.tensor([
    [0.299, 0.587, 0.114],
    [0.5969, -0.2746, -0.3213],
    [0.2115, -0.5227, 0.3112],
])

RGB_TO_LHM = torch.tensor([
    [0.2989, 0.5870, 0.1140],
    [0.3, 0.04, -0.35],
    [0.34, -0.6, 0.17],
])

RGB_TO_LMN = torch.tensor([
    [0.06, 0.63, 0.27],
    [0.30, 0.04, -0.35],
    [0.34, -0.6, 0.17],
])


_WEIGHTS = {
    ('RGB', 'YIQ'): RGB_TO_YIQ,  # HaarPSI
    ('RGB', 'Y'): RGB_TO_YIQ[:1],  # GMSD
    ('RGB', 'LHM'): RGB_TO_LHM,  # MDSI
    ('RGB', 'LMN'): RGB_TO_LMN,  # VSI
    ('RGB', 'L'): RGB_TO_LMN[:1],  # VSI
}


class ColorConv(nn.Module):
    r"""Color convolution module.

    Args:
        src: The source color space (e.g. `'RGB'`).
        dst: The destination color space (e.g. `'YIQ'`).

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> conv = ColorConv('RGB', 'YIQ')
        >>> y = conv(x)
        >>> y.size()
        torch.Size([5, 3, 256, 256])
    """

    def __init__(self, src: str, dst: str):
        super().__init__()

        assert (src, dst) in _WEIGHTS, f"Unknown {src} to {dst} conversion"

        self.register_buffer('weight', _WEIGHTS[(src, dst)])

    @property
    def device(self) -> torch.device:
        return self.weight.device

    def forward(self, x: Tensor) -> Tensor:
        return color_conv(x, self.weight)


def rgb_to_xyz(x: Tensor, value_range: float = 1.) -> Tensor:
    r"""Converts from sRGB to (CIE) XYZ.

    Wikipedia:
        https://en.wikipedia.org/wiki/SRGB
    """

    x = x / value_range

    mask = x <= 0.04045
    left = x / 12.92
    right = ((x + 0.055) / 1.055) ** 2.4

    x = torch.where(mask, left, right)

    weight = torch.tensor([
        [0.4124564, 0.3575761, 0.1804375],
        [0.2126729, 0.7151522, 0.0721750],
        [0.0193339, 0.1191920, 0.9503041],
    ])

    return color_conv(x, weight.to(x))


def xyz_to_lab(
    x: Tensor,
    illuminants: Tuple[float, float, float] = (0.956797052643698, 1., 0.9214805860173273),
) -> Tensor:
    r"""Converts from (CIE) XYZ to (CIE) LAB.

    Wikipedia:
        https://en.wikipedia.org/wiki/CIELAB_color_space
    """

    scale = torch.tensor(illuminants).view((3,) + (1,) * spatial(x))
    x = x / scale.to(x)

    delta = 6 / 29

    mask = x > delta ** 3
    left = x ** (1 / 3)
    right = x / (3 * delta ** 2) + 4 / 29

    x = torch.where(mask, left, right)

    weight = torch.tensor([
        [0., 116., 0.],
        [500., -500., 0.],
        [0., 200., -200.],
    ])

    bias = torch.tensor([-16., 0., 0.]).view(scale.shape)

    return color_conv(x, weight.to(x)) + bias.to(x)



r"""Feature Similarity (FSIM)

This module implements the FSIM in PyTorch.

Original:
    https://www4.comp.polyu.edu.hk/~cslzhang/IQA/FSIM/FSIM.htm

References:
    .. [Zhang2011] FSIM: A Feature Similarity Index for Image Quality Assessment (Zhang et al., 2011)

    .. [Kovesi1999] Image Features From Phase Congruency (Kovesi, 1999)
"""

import math
import torch
import torch.fft as fft
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from .utils import _jit, assert_type, reduce_tensor
from .utils import complex as cx
from .utils.color import ColorConv
from .utils.functional import (
    scharr_kernel,
    gradient_kernel,
    filter_grid,
    log_gabor,
    channel_conv,
    l2_norm,
)


@_jit
def fsim(
    x: Tensor,
    y: Tensor,
    pc_x: Tensor,
    pc_y: Tensor,
    kernel: Tensor,
    value_range: float = 1.,
    t1: float = 0.85,
    t2: float = 160. / (255. ** 2),
    t3: float = 200. / (255. ** 2),
    t4: float = 200. / (255. ** 2),
    lmbda: float = 0.03,
) -> Tensor:
    r"""Returns the FSIM between :math:`x` and :math:`y`,
    without color space conversion and downsampling.

    Args:
        x: An input tensor, :math:`(N, 3 \text{ or } 1, H, W)`.
        y: A target tensor, :math:`(N, 3 \text{ or } 1, H, W)`.
        pc_x: The input phase congruency, :math:`(N, H, W)`.
        pc_y: The target phase congruency, :math:`(N, H, W)`.
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
        value_range: The value range :math:`L` of the inputs (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Zhang2011]_.

    Returns:
        The FSIM vector, :math:`(N,)`.

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> y = torch.rand(5, 3, 256, 256)
        >>> filters = pc_filters(x)
        >>> pc_x = phase_congruency(x[:, :1], filters)
        >>> pc_y = phase_congruency(y[:, :1], filters)
        >>> kernel = gradient_kernel(scharr_kernel())
        >>> l = fsim(x, y, pc_x, pc_y, kernel)
        >>> l.size()
        torch.Size([5])
    """

    t2 *= value_range ** 2
    t3 *= value_range ** 2
    t4 *= value_range ** 2

    y_x, y_y = x[:, :1], y[:, :1]

    # Phase congruency similarity
    pc_m = torch.max(pc_x, pc_y)
    s_pc = (2 * pc_x * pc_y + t1) / (pc_x ** 2 + pc_y ** 2 + t1)

    # Gradient magnitude similarity
    pad = kernel.size(-1) // 2

    g_x = l2_norm(channel_conv(y_x, kernel, padding=pad), dims=[1])
    g_y = l2_norm(channel_conv(y_y, kernel, padding=pad), dims=[1])

    s_g = (2 * g_x * g_y + t2) / (g_x ** 2 + g_y ** 2 + t2)

    # Chrominance similarity
    s_l = s_pc * s_g

    if x.size(1) == 3:
        i_x, i_y = x[:, 1], y[:, 1]
        q_x, q_y = x[:, 2], y[:, 2]

        s_i = (2 * i_x * i_y + t3) / (i_x ** 2 + i_y ** 2 + t3)
        s_q = (2 * q_x * q_y + t4) / (q_x ** 2 + q_y ** 2 + t4)

        s_iq = s_i * s_q
        s_iq = cx.complx(s_iq, torch.zeros_like(s_iq))
        s_iq_lambda =  cx.real(cx.pow(s_iq, lmbda))

        s_l = s_l * s_iq_lambda

    # Feature similarity
    fs = (s_l * pc_m).sum(dim=(-1, -2)) / pc_m.sum(dim=(-1, -2))

    return fs


@_jit
def pc_filters(
    x: Tensor,
    scales: int = 4,
    orientations: int = 4,
    wavelength: float = 6.,
    factor: float = 2.,
    sigma_f: float = 0.5978,  # -log(0.55)
    sigma_theta: float = 0.6545,  # pi / (4 * 1.2)
) -> Tensor:
    r"""Returns the log-Gabor filters for :func:`phase_congruency`.

    Args:
        x: An input tensor, :math:`(*, H, W)`.
        scales: The number of scales, :math:`S_1`.
        orientations: The number of orientations, :math:`S_2`.

    Note:
        For the remaining arguments, refer to [Kovesi1999]_.

    Returns:
        The filters tensor, :math:`(S_1, S_2, H, W)`.
    """

    r, theta = filter_grid(x)

    # Low-pass filter
    lowpass = 1 / (1 + (r / 0.45) ** (2 * 15))

    # Radial
    radial = []

    for i in range(scales):
        f_0 = 1 / (wavelength * factor ** i)
        lg = log_gabor(r, f_0, sigma_f)
        radial.append(lg)

    radial = torch.stack(radial)

    # Angular
    cos_theta = torch.cos(theta)
    sin_theta = torch.sin(theta)

    theta_j = math.pi * torch.arange(orientations).to(x) / orientations
    theta_j = theta_j.view(orientations, 1, 1)

    ## Measure (theta - theta_j) in the sine/cosine domains
    ## to prevent wrap-around errors
    delta_sin = sin_theta * theta_j.cos() - cos_theta * theta_j.sin()
    delta_cos = cos_theta * theta_j.cos() + sin_theta * theta_j.sin()
    delta_theta = torch.atan2(delta_sin, delta_cos)

    angular = torch.exp(-delta_theta ** 2 / (2 * sigma_theta ** 2))

    # Combination
    filters = lowpass * radial[:, None] * angular[None, :]

    return filters


@_jit
def phase_congruency(
    x: Tensor,
    filters: Tensor,
    value_range: float = 1.,
    k: float = 2.,
    rescale: float = 1.7,
    eps: float = 1e-8,
) -> Tensor:
    r"""Returns the Phase Congruency (PC) of :math:`x`.

    Args:
        x: An input tensor, :math:`(N, 1, H, W)`.
        filters: The frequency domain filters, :math:`(S_1, S_2, H, W)`.
        value_range: The value range :math:`L` of the input (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Kovesi1999]_.

    Returns:
        The PC tensor, :math:`(N, H, W)`.

    Example:
        >>> x = torch.rand(5, 1, 256, 256)
        >>> filters = pc_filters(x)
        >>> pc = phase_congruency(x, filters)
        >>> pc.size()
        torch.Size([5, 256, 256])
    """

    x = x * (255. / value_range)

    # Filters
    M_hat = filters
    M = fft.ifft2(M_hat)
    M = cx.real(torch.view_as_real(M))

    # Even & odd (real and imaginary) responses
    eo = fft.ifft2(fft.fft2(x[:, None]) * M_hat)
    eo = torch.view_as_real(eo)

    # Amplitude
    A = cx.mod(eo)

    # Expected E^2
    A2 = A[:, 0] ** 2
    median_A2, _ = A2.flatten(-2).median(dim=-1)
    expect_A2 = median_A2 / math.log(2)

    expect_M2_hat = (M_hat[0] ** 2).mean(dim=(-1, -2))
    expect_MiMj = (M[:, None] * M[None, :]).sum(dim=(0, 1, 3, 4))

    expect_E2 = expect_A2 * expect_MiMj / expect_M2_hat

    # Threshold
    sigma_G = expect_E2.sqrt()
    mu_R = sigma_G * (math.pi / 2) ** 0.5
    sigma_R = sigma_G * (2 - math.pi / 2) ** 0.5

    T = mu_R + k * sigma_R
    T = T / rescale  # emprirical rescaling
    T = T[..., None, None]

    # Phase deviation
    FH = eo.sum(dim=1, keepdim=True)
    phi_eo = FH / (cx.mod(FH)[..., None] + eps)

    E = cx.dot(eo, phi_eo) - cx.dot(eo, cx.turn(phi_eo)).abs()
    E = E.sum(dim=1)

    # Phase congruency
    pc = (E - T).relu().sum(dim=1) / (A.sum(dim=(1, 2)) + eps)

    return pc


class FSIM(nn.Module):
    r"""Creates a criterion that measures the FSIM
    between an input and a target.

    Before applying :func:`fsim`, the input and target are converted from
    RBG to Y(IQ) and downsampled by a factor :math:`\frac{\min(H, W)}{256}`.

    Args:
        chromatic: Whether to use the chromatic channels (IQ) or not.
        downsample: Whether downsampling is enabled or not.
        kernel: A gradient kernel, :math:`(2, 1, K, K)`.
            If `None`, use the Scharr kernel instead.
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Note:
        `**kwargs` are passed to :func:`fsim`.

    Shapes:
        input: :math:`(N, 3, H, W)`
        target: :math:`(N, 3, H, W)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = FSIM().cuda()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = 1 - criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    def __init__(
        self,
        chromatic: bool = True,
        downsample: bool = True,
        kernel: Tensor = None,
        reduction: str = 'mean',
        **kwargs,
    ):
        super().__init__()

        if kernel is None:
            kernel = gradient_kernel(scharr_kernel())

        self.register_buffer('kernel', kernel)
        self.register_buffer('filters', torch.zeros((0, 0, 0, 0)))

        self.convert = ColorConv('RGB', 'YIQ' if chromatic else 'Y')
        self.downsample = downsample
        self.reduction = reduction
        self.value_range = kwargs.get('value_range', 1.)
        self.kwargs = kwargs

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            device=self.kernel.device,
            dim_range=(4, 4),
            n_channels=3,
            value_range=(0., self.value_range),
        )

        # Downsample
        if self.downsample:
            _, _, h, w = input.size()
            M = round(min(h, w) / 256)

            if M > 1:
                input = F.avg_pool2d(input, kernel_size=M, ceil_mode=True)
                target = F.avg_pool2d(target, kernel_size=M, ceil_mode=True)

        # RGB to Y(IQ)
        input = self.convert(input)
        target = self.convert(target)

        # Phase congruency
        if self.filters.shape[-2:] != input.shape[-2:]:
            self.filters = pc_filters(input)

        pc_input = phase_congruency(input[:, :1], self.filters, self.value_range)
        pc_target = phase_congruency(target[:, :1], self.filters, self.value_range)

        # FSIM
        l = fsim(input, target, pc_input, pc_target, kernel=self.kernel, **self.kwargs)

        return reduce_tensor(l, self.reduction)



r"""Learned Perceptual Image Patch Similarity (LPIPS)

This module implements the LPIPS in PyTorch.

Original:
    https://github.com/richzhang/PerceptualSimilarity

References:
    .. [Zhang2018] The Unreasonable Effectiveness of Deep Features as a Perceptual Metric (Zhang et al., 2018)

    .. [Deng2009] ImageNet: A large-scale hierarchical image database (Deng et al, 2009)
"""

import inspect
import os
import torch
import torch.nn as nn
import torchvision.models as models
import torch.hub as hub

from torch import Tensor
from typing import Dict, List

from .utils import _jit, assert_type, reduce_tensor
from .utils.functional import l2_norm


ORIGIN: str = 'https://github.com/richzhang/PerceptualSimilarity'
SHIFT: Tensor = torch.tensor([0.485, 0.456, 0.406])
SCALE: Tensor = torch.tensor([0.229, 0.224, 0.225])


def get_weights(
    network: str = 'alex',
    version: str = 'v0.1',
) -> Dict[str, Tensor]:
    r"""Returns the official LPIPS weights for `network`.

    Args:
        network: Specifies the perception network that is used:
            `'alex'` | `'squeeze'` | `'vgg'`.
        version: Specifies the official version release:
            `'v0.0'` | `'v0.1'`.

    Example:
        >>> w = get_weights(network='alex')
        >>> w.keys()
        dict_keys(['0.1.weight', '1.1.weight', '2.1.weight', '3.1.weight', '4.1.weight'])
    """

    # Load from URL
    weights = hub.load_state_dict_from_url(
        f'{ORIGIN}/raw/master/lpips/weights/{version}/{network}.pth',
        map_location='cpu',
    )

    # Format keys
    weights = {
        k.replace('lin', '').replace('.model', ''): v
        for (k, v) in weights.items()
    }

    return weights


class Intermediary(nn.Module):
    r"""Module that catches and returns the outputs of indermediate
    target layers of a sequential module during its forward pass.

    Args:
        layers: A sequential module.
        targets: A list of target layer indexes.
    """

    def __init__(self, layers: nn.Sequential, targets: List[int]):
        super().__init__()

        self.layers = nn.ModuleList()
        j = 0

        seq: List[nn.Module] = []

        for i, layer in enumerate(layers):
            seq.append(layer)

            if i == targets[j]:
                self.layers.append(nn.Sequential(*seq))
                seq.clear()

                j += 1
                if j == len(targets):
                    break

    def forward(self, input: Tensor) -> List[Tensor]:
        output = []

        for layer in self.layers:
            input = layer(input)
            output.append(input)

        return output


class LPIPS(nn.Module):
    r"""Creates a criterion that measures the LPIPS
    between an input :math:`x` and a target :math:`y`.

    .. math::
        \text{LPIPS}(x, y) = \sum_{l \, \in \, \mathcal{F}}
            w_l \cdot \text{MSE}(\phi_l(x), \phi_l(y))

    where :math:`\phi_l` represents the normalized output of an intermediate
    layer :math:`l` in a perceptual network :math:`\mathcal{F}`.

    Note:
        :class:`LPIPS` is a trainable metric. For more details, refer to [Zhang2018]_.

    Args:
        network: Specifies the perceptual network :math:`\mathcal{F}` to use:
            `'alex'` | `'squeeze'` | `'vgg'`.
        scaling: Whether the input and target need to be scaled w.r.t. [Deng2009]_.
        dropout: Whether dropout is used or not.
        pretrained: Whether the official weights :math:`w_l` are used or not.
        eval: Whether to initialize the object in evaluation mode or not.
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Shapes:
        input: :math:`(N, 3, H, W)`
        target: :math:`(N, 3, H, W)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = LPIPS().cuda()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    def __init__(
        self,
        network: str = 'alex',
        scaling: bool = True,
        dropout: bool = False,
        pretrained: bool = True,
        eval: bool = True,
        reduction: str = 'mean',
    ):
        super().__init__()

        # ImageNet scaling
        self.scaling = scaling
        self.register_buffer('shift', SHIFT.view(1, -1, 1, 1))
        self.register_buffer('scale', SCALE.view(1, -1, 1, 1))

        # Perception layers
        if network == 'alex':  # AlexNet
            layers = models.alexnet(pretrained=True).features
            targets = [1, 4, 7, 9, 11]
            channels = [64, 192, 384, 256, 256]
        elif network == 'squeeze':  # SqueezeNet
            layers = models.squeezenet1_1(pretrained=True).features
            targets = [1, 4, 7, 9, 10, 11, 12]
            channels = [64, 128, 256, 384, 384, 512, 512]
        elif network == 'vgg':  # VGG16
            layers = models.vgg16(pretrained=True).features
            targets = [3, 8, 15, 22, 29]
            channels = [64, 128, 256, 512, 512]
        else:
            raise ValueError(f'Unknown network architecture {network}')

        self.net = Intermediary(layers, targets)
        for p in self.net.parameters():
            p.requires_grad = False

        # Linear comparators
        self.lins = nn.ModuleList([
            nn.Sequential(
                nn.Dropout(inplace=True) if dropout else nn.Identity(),
                nn.Conv2d(c, 1, kernel_size=1, bias=False),
            ) for c in channels
        ])

        if pretrained:
            self.lins.load_state_dict(get_weights(network=network))

        if eval:
            self.eval()

        self.reduction = reduction

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            device=self.shift.device,
            dim_range=(4, 4),
            n_channels=3,
            value_range=(0., 1.) if self.scaling else (0., -1.),
        )

        # ImageNet scaling
        if self.scaling:
            input = (input - self.shift) / self.scale
            target = (target - self.shift) / self.scale

        # LPIPS
        residuals = []

        for lin, fx, fy in zip(self.lins, self.net(input), self.net(target)):
            fx = fx / l2_norm(fx, dims=[1], keepdim=True)
            fy = fy / l2_norm(fy, dims=[1], keepdim=True)

            mse = ((fx - fy) ** 2).mean(dim=(-1, -2), keepdim=True)
            residuals.append(lin(mse).flatten())

        l = torch.stack(residuals, dim=-1).sum(dim=-1)

        return reduce_tensor(l, self.reduction)



r"""General purpose tensor functionals"""

import torch
import torch.fft as fft
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import List, Tuple, Union


def channel_conv(
    x: Tensor,
    kernel: Tensor,
    padding: int = 0,  # Union[int, Tuple[int, ...]]
) -> Tensor:
    r"""Returns the channel-wise convolution of :math:`x` with the kernel `kernel`.

    Args:
        x: A tensor, :math:`(N, C, *)`.
        kernel: A kernel, :math:`(C', 1, *)`.
        padding: The implicit paddings on both sides of the input dimensions.

    Example:
        >>> x = torch.arange(25, dtype=torch.float).view(1, 1, 5, 5)
        >>> x
        tensor([[[[ 0.,  1.,  2.,  3.,  4.],
                  [ 5.,  6.,  7.,  8.,  9.],
                  [10., 11., 12., 13., 14.],
                  [15., 16., 17., 18., 19.],
                  [20., 21., 22., 23., 24.]]]])
        >>> kernel = torch.ones((1, 1, 3, 3))
        >>> channel_conv(x, kernel)
        tensor([[[[ 54.,  63.,  72.],
                  [ 99., 108., 117.],
                  [144., 153., 162.]]]])
    """

    D = len(kernel.shape) - 2

    assert D <= 3, "PyTorch only supports 1D, 2D or 3D convolutions."

    if D == 3:
        return F.conv3d(x, kernel, padding=padding, groups=x.size(-4))
    elif D == 2:
        return F.conv2d(x, kernel, padding=padding, groups=x.size(-3))
    elif D == 1:
        return F.conv1d(x, kernel, padding=padding, groups=x.size(-2))
    else:
        return F.linear(x, kernel.expand(x.size(-1)))


def channel_convs(
    x: Tensor,
    kernels: List[Tensor],
    padding: int = 0,  # Union[int, Tuple[int, ...]]
) -> Tensor:
    r"""Returns the channel-wise convolution of :math:`x` with
    the series of kernel `kernels`.

    Args:
        x: A tensor, :math:`(N, C, *)`.
        kernels: A list of kernels, each :math:`(C', 1, *)`.
        padding: The implicit paddings on both sides of the input dimensions.

    Example:
        >>> x = torch.arange(25, dtype=torch.float).view(1, 1, 5, 5)
        >>> x
        tensor([[[[ 0.,  1.,  2.,  3.,  4.],
                  [ 5.,  6.,  7.,  8.,  9.],
                  [10., 11., 12., 13., 14.],
                  [15., 16., 17., 18., 19.],
                  [20., 21., 22., 23., 24.]]]])
        >>> kernels = [torch.ones((1, 1, 3, 1)), torch.ones((1, 1, 1, 3))]
        >>> channel_convs(x, kernels)
        tensor([[[[ 54.,  63.,  72.],
                  [ 99., 108., 117.],
                  [144., 153., 162.]]]])
    """

    if padding > 0:
        pad = (padding,) * (2 * x.dim() - 4)
        x = F.pad(x, pad=pad)

    for k in kernels:
        x = channel_conv(x, k)

    return x


def gaussian_kernel(
    size: int,
    sigma: float = 1.
) -> Tensor:
    r"""Returns the 1-dimensional Gaussian kernel of size :math:`K`.

    .. math::
        G(x) = \gamma \exp \left(\frac{(x - \mu)^2}{2 \sigma^2}\right)

    where :math:`\gamma` is such that

    .. math:: \sum_{x = 1}^{K} G(x) = 1

    and :math:`\mu = \frac{1 + K}{2}`.

    Args:
        size: The kernel size :math:`K`.
        sigma: The standard deviation :math:`\sigma` of the distribution.

    Returns:
        The kernel vector, :math:`(K,)`.

    Note:
        An :math:`N`-dimensional Gaussian kernel is separable, meaning that
        applying it is equivalent to applying a series of :math:`N` 1-dimensional
        Gaussian kernels, which has a lower computational complexity.

    Wikipedia:
        https://en.wikipedia.org/wiki/Gaussian_blur

    Example:
        >>> gaussian_kernel(5, sigma=1.5)
        tensor([0.1201, 0.2339, 0.2921, 0.2339, 0.1201])
    """

    kernel = torch.arange(size, dtype=torch.float)
    kernel -= (size - 1) / 2
    kernel = kernel ** 2 / (2. * sigma ** 2)
    kernel = torch.exp(-kernel)
    kernel /= kernel.sum()

    return kernel


def kernel_views(kernel: Tensor, n: int = 2) -> List[Tensor]:
    r"""Returns the :math:`N`-dimensional views of the 1-dimensional
    kernel `kernel`.

    Args:
        kernel: A kernel, :math:`(C, 1, K)`.
        n: The number of dimensions :math:`N`.

    Returns:
        The list of views, each :math:`(C, 1, \underbrace{1, \dots, 1}_{i}, K, \underbrace{1, \dots, 1}_{N - i - 1})`.

    Example:
        >>> kernel = gaussian_kernel(5, sigma=1.5).repeat(3, 1, 1)
        >>> kernel.size()
        torch.Size([3, 1, 5])
        >>> views = kernel_views(kernel, n=2)
        >>> views[0].size(), views[1].size()
        (torch.Size([3, 1, 5, 1]), torch.Size([3, 1, 1, 5]))
    """

    if n == 1:
        return [kernel]
    elif n == 2:
        return [kernel.unsqueeze(-1), kernel.unsqueeze(-2)]

    # elif n > 2:
    c, _, k = kernel.size()

    shape: List[int] = [c, 1] + [1] * n
    views = []

    for i in range(2, n + 2):
        shape[i] = k
        views.append(kernel.view(shape))
        shape[i] = 1

    return views


def haar_kernel(size: int) -> Tensor:
    r"""Returns the horizontal Haar kernel.

    Args:
        size: The kernel (even) size :math:`K`.

    Returns:
        The kernel, :math:`(K, K)`.

    Wikipedia:
        https://en.wikipedia.org/wiki/Haar_wavelet

    Example:
        >>> haar_kernel(2)
        tensor([[ 0.5000, -0.5000],
                [ 0.5000, -0.5000]])
    """

    return torch.outer(
        torch.ones(size) / size,
        torch.tensor([1., -1.]).repeat_interleave(size // 2),
    )


def prewitt_kernel() -> Tensor:
    r"""Returns the Prewitt kernel.

    Returns:
        The kernel, :math:`(3, 3)`.

    Wikipedia:
        https://en.wikipedia.org/wiki/Prewitt_operator

    Example:
        >>> prewitt_kernel()
        tensor([[ 0.3333,  0.0000, -0.3333],
                [ 0.3333,  0.0000, -0.3333],
                [ 0.3333,  0.0000, -0.3333]])
    """

    return torch.outer(
        torch.tensor([1., 1., 1.]) / 3,
        torch.tensor([1., 0., -1.]),
    )


def sobel_kernel() -> Tensor:
    r"""Returns the Sobel kernel.

    Returns:
        The kernel, :math:`(3, 3)`.

    Wikipedia:
        https://en.wikipedia.org/wiki/Sobel_operator

    Example:
        >>> sobel_kernel()
        tensor([[ 0.2500,  0.0000, -0.2500],
                [ 0.5000,  0.0000, -0.5000],
                [ 0.2500,  0.0000, -0.2500]])
    """

    return torch.outer(
        torch.tensor([1., 2., 1.]) / 4,
        torch.tensor([1., 0., -1.]),
    )


def scharr_kernel() -> Tensor:
    r"""Returns the Scharr kernel.

    Returns:
        The kernel, :math:`(3, 3)`.

    Wikipedia:
        https://en.wikipedia.org/wiki/Scharr_operator

    Example:
        >>> scharr_kernel()
        tensor([[ 0.1875,  0.0000, -0.1875],
                [ 0.6250,  0.0000, -0.6250],
                [ 0.1875,  0.0000, -0.1875]])
    """

    return torch.outer(
        torch.tensor([3., 10., 3.]) / 16,
        torch.tensor([1., 0., -1.]),
    )


def gradient_kernel(kernel: Tensor) -> Tensor:
    r"""Returns `kernel` transformed into a gradient.

    Args:
        kernel: A convolution kernel, :math:`(K, K)`.

    Returns:
        The gradient kernel, :math:`(2, 1, K, K)`.

    Example:
        >>> g = gradient_kernel(prewitt_kernel())
        >>> g.size()
        torch.Size([2, 1, 3, 3])
    """

    return torch.stack([kernel, kernel.t()]).unsqueeze(1)


def filter_grid(x: Tensor) -> Tuple[Tensor, Tensor]:
    r"""Returns the (quadrant-shifted) frequency grid for :math:`x`.

    Args:
        x: An input tensor, :math:`(*, H, W)`.

    Returns:
        The radius and phase tensors, both :math:`(H, W)`.

    Example:
        >>> x = torch.rand(5, 5)
        >>> r, phi = filter_grid(x)
        >>> r
        tensor([[0.0000, 0.2500, 0.5000, 0.5000, 0.2500],
                [0.2500, 0.3536, 0.5590, 0.5590, 0.3536],
                [0.5000, 0.5590, 0.7071, 0.7071, 0.5590],
                [0.5000, 0.5590, 0.7071, 0.7071, 0.5590],
                [0.2500, 0.3536, 0.5590, 0.5590, 0.3536]])
        >>> phi
        tensor([[-0.0000, -1.5708, -1.5708,  1.5708,  1.5708],
                [-0.0000, -0.7854, -1.1071,  1.1071,  0.7854],
                [-0.0000, -0.4636, -0.7854,  0.7854,  0.4636],
                [-3.1416, -2.6779, -2.3562,  2.3562,  2.6779],
                [-3.1416, -2.3562, -2.0344,  2.0344,  2.3562]])
    """

    u, v = [
        (torch.arange(n).to(x) - n // 2) / (n - n % 2)
        for n in x.shape[-2:]
    ]
    u, v = fft.ifftshift(u[:, None]), fft.ifftshift(v[None, :])

    r = (u ** 2 + v ** 2).sqrt()
    phi = torch.atan2(-v, u)

    return r, phi


def log_gabor(f: Tensor, f_0: float, sigma_f: float) -> Tensor:
    r"""Returns the log-Gabor filter of :math:`f`.

    .. math::
        G(f) = \exp \left( - \frac{\log(f / f_0)^2}{2 \sigma_f^2} \right)

    Args:
        f: A frequency tensor, :math:`(*,)`.
        f_0: The center frequency :math:`f_0`.
        sigma_f: The bandwidth (log-)deviation :math:`\sigma_f`.

    Returns:
        The filter tensor, :math:`(*,)`.

    Wikipedia:
        https://en.wikipedia.org/wiki/Log_Gabor_filter

    Example:
        >>> x = torch.rand(5, 5)
        >>> r, phi = filter_grid(x)
        >>> log_gabor(r, 1., 1.)
        tensor([[0.0000, 0.3825, 0.7864, 0.7864, 0.3825],
                [0.3825, 0.5825, 0.8444, 0.8444, 0.5825],
                [0.7864, 0.8444, 0.9417, 0.9417, 0.8444],
                [0.7864, 0.8444, 0.9417, 0.9417, 0.8444],
                [0.3825, 0.5825, 0.8444, 0.8444, 0.5825]])
    """

    return torch.exp(- (f / f_0).log() ** 2 / (2 * sigma_f ** 2))


def l2_norm(
    x: torch.Tensor,
    dims: List[int],
    keepdim: bool = False,
) -> torch.Tensor:
    r"""Returns the :math:`L_2` norm of :math:`x`.

    .. math:
        L_2(x) = \left\| x \right\|_2 = \sqrt{\sum_i x^2_i}

    Args:
        x: A tensor, :math:`(*,)`.
        dims: The dimensions along which to calculate the norm.
        keepdim: Whether the output tensor has `dims` retained or not.

    Wikipedia:
        https://en.wikipedia.org/wiki/Norm_(mathematics)

    Example:
        >>> x = torch.arange(9).float().view(3, 3)
        >>> x
        tensor([[0., 1., 2.],
                [3., 4., 5.],
                [6., 7., 8.]])
        >>> l2_norm(x, dims=[0])
        tensor([6.7082, 8.1240, 9.6437])
    """

    x = x ** 2
    x = x.sum(dim=dims, keepdim=keepdim)
    x = x.sqrt()

    return x



r"""Haar Perceptual Similarity Index (HaarPSI)

This module implements the HaarPSI in PyTorch.

Original:
    https://github.com/rgcda/haarpsi

Wikipedia:
    https://en.wikipedia.org/wiki/Haar_wavelet

References:
    .. [Reisenhofer2018] A Haar Wavelet-Based Perceptual Similarity Index for Image Quality Assessment (Reisenhofer et al., 2018)
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from .utils import _jit, assert_type, reduce_tensor
from .utils.color import ColorConv
from .utils.functional import (
    haar_kernel,
    gradient_kernel,
    channel_conv,
)


@_jit
def haarpsi(
    x: Tensor,
    y: Tensor,
    n_kernels: int = 3,
    value_range: float = 1.,
    c: float = 30. / (255. ** 2),
    alpha: float = 4.2,
) -> Tensor:
    r"""Returns the HaarPSI between :math:`x` and :math:`y`,
    without color space conversion.

    Args:
        x: An input tensor, :math:`(N, 3 \text{ or } 1, H, W)`.
        y: A target tensor, :math:`(N, 3 \text{ or } 1, H, W)`.
        n_kernels: The number of Haar wavelet kernels to use.
        value_range: The value range :math:`L` of the inputs (usually `1.` or `255`).

    Note:
        For the remaining arguments, refer to [Reisenhofer2018]_.

    Returns:
        The HaarPSI vector, :math:`(N,)`.

    Example:
        >>> x = torch.rand(5, 3, 256, 256)
        >>> y = torch.rand(5, 3, 256, 256)
        >>> l = haarpsi(x, y)
        >>> l.size()
        torch.Size([5])
    """

    c *= value_range ** 2

    # Y
    y_x, y_y = x[:, :1], y[:, :1]

    ## Gradient(s)
    g_xy: List[Tuple[Tensor, Tensor]] = []

    for j in range(1, n_kernels + 1):
        kernel_size = int(2 ** j)

        ### Haar wavelet kernel
        kernel = gradient_kernel(haar_kernel(kernel_size)).to(x.device)

        ### Haar filter (gradient)
        pad = kernel_size // 2

        g_x = channel_conv(y_x, kernel, padding=pad)[..., 1:, 1:].abs()
        g_y = channel_conv(y_y, kernel, padding=pad)[..., 1:, 1:].abs()

        g_xy.append((g_x, g_y))

    ## Gradient similarity(ies)
    gs = []
    for g_x, g_y in g_xy[:-1]:
        gs.append((2. * g_x * g_y + c) / (g_x ** 2 + g_y ** 2 + c))

    ## Local similarity(ies)
    ls = torch.stack(gs, dim=-1).sum(dim=-1) / 2.  # (N, 2, H, W)

    ## Weight(s)
    w = torch.stack(g_xy[-1], dim=-1).max(dim=-1)[0]  # (N, 2, H, W)

    # IQ
    if x.size(1) == 3:
        iq_x, iq_y = x[:, 1:], y[:, 1:]

        ## Mean filter
        m_x = F.avg_pool2d(iq_x, 2, stride=1, padding=1)[..., 1:, 1:].abs()
        m_y = F.avg_pool2d(iq_y, 2, stride=1, padding=1)[..., 1:, 1:].abs()

        ## Chromatic similarity(ies)
        cs = (2. * m_x * m_y + c) / (m_x ** 2 + m_y ** 2 + c)

        ## Local similarity(ies)
        ls = torch.cat([ls, cs.mean(1, True)], dim=1)  # (N, 3, H, W)

        ## Weight(s)
        w = torch.cat([w, w.mean(1, True)], dim=1)  # (N, 3, H, W)

    # HaarPSI
    hs = torch.sigmoid(ls * alpha)
    hpsi = (hs * w).sum(dim=(-1, -2, -3)) / w.sum(dim=(-1, -2, -3))
    hpsi = (torch.logit(hpsi) / alpha) ** 2

    return hpsi


class HaarPSI(nn.Module):
    r"""Creates a criterion that measures the HaarPSI
    between an input and a target.

    Before applying :func:`haarpsi`, the input and target are converted from
    RBG to Y(IQ) and downsampled by a factor 2.

    Args:
        chromatic: Whether to use the chromatic channels (IQ) or not.
        downsample: Whether downsampling is enabled or not.
        reduction: Specifies the reduction to apply to the output:
            `'none'` | `'mean'` | `'sum'`.

    Note:
        `**kwargs` are passed to :func:`haarpsi`.

    Shapes:
        input: :math:`(N, 3, H, W)`
        target: :math:`(N, 3, H, W)`
        output: :math:`(N,)` or :math:`()` depending on `reduction`

    Example:
        >>> criterion = HaarPSI().cuda()
        >>> x = torch.rand(5, 3, 256, 256, requires_grad=True).cuda()
        >>> y = torch.rand(5, 3, 256, 256).cuda()
        >>> l = 1 - criterion(x, y)
        >>> l.size()
        torch.Size([])
        >>> l.backward()
    """

    def __init__(
        self,
        chromatic: bool = True,
        downsample: bool = True,
        reduction: str = 'mean',
        **kwargs,
    ):
        super().__init__()

        self.convert = ColorConv('RGB', 'YIQ' if chromatic else 'Y')
        self.downsample = downsample
        self.reduction = reduction
        self.value_range = kwargs.get('value_range', 1.)
        self.kwargs = kwargs

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        assert_type(
            input, target,
            device=self.convert.device,
            dim_range=(4, 4),
            n_channels=3,
            value_range=(0., self.value_range),
        )

        # Downsample
        if self.downsample:
            input = F.avg_pool2d(input, 2, ceil_mode=True)
            target = F.avg_pool2d(target, 2, ceil_mode=True)

        # RGB to Y(IQ)
        input = self.convert(input)
        target = self.convert(target)

        # HaarPSI
        l = haarpsi(input, target, **self.kwargs)

        return reduce_tensor(l, self.reduction)



r"""PyTorch Image Quality Assessement (PIQA)

The :mod:`piqa` package is divided in several submodules, each of
which implements the functions and/or classes related to a
specific image quality assessement metric.
"""

from .tv import TV
from .psnr import PSNR
from .ssim import SSIM, MS_SSIM
from .lpips import LPIPS
from .gmsd import GMSD, MS_GMSD
from .mdsi import MDSI
from .haarpsi import HaarPSI
from .vsi import VSI
from .fsim import FSIM
